/******************************************************************************
Author:         David Szybala
Website:       www.davidszybala.com
Alt Website:  www.twistedgerbil.com
Program:       rename_materials.ms
Date:           2/10/09
Description:   This script provides access to special renaming functions for materials
******************************************************************************/

try (destroyDialog rename_material_rollout) catch()


struct RenameMaterials
(
    materials_all = #(),
    materials_filtered = #(),
    material_children = #(),
    material_previews = #(),
    
    /**************************************************************************
    Sorts items by name
    **************************************************************************/
    fn sort_by_name item_one item_two =
    (
        case of
        (
            (item_one.name < item_two.name): -1
            (item_one.name > item_two.name): 1
            default: 0
        ) --end case
    ), --end fn
    
    /**************************************************************************
    Returns all root materials in the scene sorted by name
    **************************************************************************/
    fn get_root_materials =
    (
        local roots = for current in sceneMaterials where (superClassOf current) == material collect current
        
        if roots.count > 0 do
            qsort roots this.sort_by_name
        
        roots --return value
    ), --end fn
    
    /**************************************************************************
    Recursive function. Collects all the sub materials and their names from the
    argument 'parent_material' WARNING: Check each child against all children
    of the parent since some materials (VRayBlendMtl for example) will report
    the same child multiple times
    **************************************************************************/
    fn collect_sub_materials parent_material =
    (
        for x = 1 to (getNumSubMtls parent_material) do
        (
            child_material = getSubMtl parent_material x
            
            if child_material != undefined do
            (
                if (findItem this.material_children child_material) == 0 do
                (
                    append this.materials_all child_material
                    append this.material_children child_material
                    
                    if (getNumSubMtls child_material) > 0 do
                        collect_sub_materials child_material
                ) --end if
            ) --end if
        ) --end for
    ), --end fn
    
    /**************************************************************************
    Filters the current scene materials by name
    **************************************************************************/
    fn filter_materials filter_text =
    (
        --CRITICAL! Clear previous data
        free this.materials_filtered
        
        if filter_text != "" then
        (
            for current in this.materials_all do
            (
                if (matchPattern current.name pattern:("*" + filter_text + "*") ignorecase:false) do
                    append this.materials_filtered current
            ) --end if
        ) --end if
        else
            this.materials_filtered = deepcopy this.materials_all
    ), --end fn
    
    /**************************************************************************
    Collects a list of all materials and their names in the current scene
    **************************************************************************/
    fn collect_materials include_subs filter_text =
    (
        --CRITICAL! Clear previous data
        free this.materials_all
        
        for current in (this.get_root_materials()) do
        (
            append this.materials_all current
        
            if include_subs do
            (
                if (getNumSubMtls current) > 0 do
                (
                    --CRITICAL! Clear previous data!
                    free this.material_children
                    
                    collect_sub_materials current
                ) --end if
            ) --end if
        ) --end for
        
        if include_subs do
            qsort this.materials_all this.sort_by_name
        
        this.filter_materials filter_text
    ), --end fn
    
    /**************************************************************************
    Performs a remove first/last # and returns the result
    **************************************************************************/
    fn get_remove current_name start_index end_index =
    (
        current_name = substring current_name (start_index + 1) (current_name.count - (start_index + end_index))
        current_name --return value
    ), --end fn
    
    /**************************************************************************
    Performs a case sensitive find and replace and returns the result
    **************************************************************************/
    fn get_case_replace current_name find_text replace_text first_only =
    (
        local index = 1
        local continue_search = true
        
        if (matchPattern current_name pattern:("*" + find_text + "*") ignorecase:false) do
        (
            if first_only then
            (
                while continue_search do
                (
                    if current_name[index] == find_text[1] do
                    (
                        if (substring current_name index find_text.count) == find_text do
                        (
                            current_name = replace current_name index find_text.count replace_text
                            continue_search = false
                        ) --end if
                    ) --end if
                    
                    index += 1
                ) --end while
            ) --end if
            else
                current_name = substituteString current_name find_text replace_text
        ) --end if
        
        current_name --return value        
    ), --end fn 
    
    /**************************************************************************
    Performs a case insensitive find and replace and returns the result
    **************************************************************************/
    fn get_replace current_name find_text replace_text first_only =
    (
        local index = 1
        
        if (matchPattern current_name pattern:("*" + find_text + "*") ignorecase:true) do
        (
            index = findString current_name find_text
            
            while index != undefined do
            (
                current_name = replace current_name index find_text.count replace_text
                
                if first_only then
                    index = undefined
                else
                    index = findString current_name find_text
            ) --end while
        ) --end if
        
        current_name --return value
    ), --end fn
    
     /**************************************************************************
    Performs a name edit and returns the result
    **************************************************************************/
    fn get_edit current_name user_prefix user_base user_suffix =
    (
        if user_base != "" do
            current_name = user_base
        
        current_name = user_prefix + current_name
        current_name = current_name + user_suffix
            
        current_name --return value
    ), --end fn
    
    /**************************************************************************
    Returns a numeric padding string based on 'counter'
    **************************************************************************/
    fn get_numeric_padding counter user_padding =
    (
        local total_padding = 0
        local result = ""
        
        total_padding = user_padding - ((counter as string).count)
        result = counter as string
        
        for i = 1 to total_padding do
            result = "0" + result
        
        result --return value
    ), --end fn
    
    /**************************************************************************
    Returns an alphabetic padding string based on 'counter'
    **************************************************************************/
    fn get_alpha_padding counter =
    (
        local total_padding = 0
        local current_letter = ""
        local result = ""
        
        total_padding = (counter / 26) as integer --will return 0 if counter < 26
        current_letter = bit.intAsChar (counter - (26 * total_padding) + 65)
        result = current_letter
        
        for i = 1 to total_padding do
            result = current_letter + result
        
        result --return value
    ), --end fn
    
    /**************************************************************************
    Performs a casing operation and returns the result
    **************************************************************************/
    fn get_case current_name case_state =
    (
        case case_state of
        (
            (2): current_name = toUpper current_name
            (3): current_name = toLower current_name
        ) --end case
        
        current_name --return value
    ), --end fn
    
    /**************************************************************************
    Shortens the name if longer than 'max_length' and returns the result
    **************************************************************************/
    fn get_shorten current_name max_length =
    (
        if current_name.count > max_length do
            current_name = substring current_name 1 max_length
            
        current_name --return value
    ), --end fn
    
    /**************************************************************************
    Gets a unique version of 'current_name' and returns the result
    **************************************************************************/
    fn get_unique current_name material_index =
    (
        local current_names = for current in this.materials_all collect current.name
        local unique_name = current_name
        local counter = 0
        
        --replace current names with preview names, but only to the current material
        for i = 1 to material_index do
        (
            index = findItem this.materials_all this.materials_filtered[i]
            current_names[index] = this.material_previews[i]
        ) --end for
        
        --remove the current material name from the list
        index = findItem this.materials_all this.materials_filtered[material_index]
        current_names[index] = ""
        
        --compare against remaining names
        while (findItem current_names unique_name) > 0 do
        (
            counter += 1
            unique_name = (current_name + "_" + (counter as string))
        ) --end while
            
        unique_name --return value
    ) --end fn
) --end struct


( --put script into local scope
    global rename_material_rollout
    local rename_material_ini = getDir #plugcfg + "\\RenameMaterials.ini"
    local rename_material_position
    
    rollout rename_material_rollout "Rename Materials"
    (
        local renamer = RenameMaterials()
        
        /**********************************************************************
        Refreshes the material list
        **********************************************************************/
        fn refresh_list =
        (
            renamer.collect_materials rename_material_rollout.include_sub_checkbox.checked rename_material_rollout.filter_text.text
            rename_material_rollout.material_list.items = for current in renamer.materials_filtered collect current.name
        ) --end fn
        
        /**********************************************************************
        Resets the remove ui
        **********************************************************************/
        fn reset_remove_ui =
        (
            rename_material_rollout.remove_first_spinner.value = 0
            rename_material_rollout.remove_last_spinner.value = 0
        ) --end fn
        
        /**********************************************************************
        Resets the find/replace ui
        **********************************************************************/
        fn reset_find_ui =
        (
            rename_material_rollout.find_text.text = ""
            rename_material_rollout.replace_text.text = ""
            rename_material_rollout.first_only_checkbox.checked = false
            rename_material_rollout.match_case_checkbox.checked = true
        ) --end fn
        
        /**********************************************************************
        Resets the edit name ui
        **********************************************************************/
        fn reset_edit_ui =
        (
            rename_material_rollout.prefix_text.text = ""
            rename_material_rollout.base_text.text = ""
            rename_material_rollout.suffix_text.text = ""
        ) --end fn
        
        /**********************************************************************
        Resets the increment ui
        **********************************************************************/
        fn reset_increment_ui =
        (
            rename_material_rollout.pad_mode_radio.state = 3
            rename_material_rollout.padding_spinner.value = 1
            rename_material_rollout.padding_spinner.enabled = false
        ) --end fn
        
        /**********************************************************************
        Resets the misc ui
        **********************************************************************/
        fn reset_misc_ui =
        (
            rename_material_rollout.case_radio.state = 1
            rename_material_rollout.name_length_checkbox.checked = false
            rename_material_rollout.name_length_spinner.value = 60
        ) --end fn
    
        /**********************************************************************
        UI creation
        **********************************************************************/
        groupbox remove_group "1. Remove" pos:[5,5] width:215 height:40
            spinner remove_first_spinner "First:" pos:[15,22] range:[0,999,0] fieldWidth:30 type:#integer scale:1
            spinner remove_last_spinner "Last:" pos:[90,22] range:[0,999,0] fieldWidth:30 type:#integer scale:1
        
        groupbox find_group "2. Find and Replace" pos:[5,50] width:215 height:100
            label find_label "Find:" pos:[15,70] width:30 height:15
            edittext find_text "" pos:[60,68] fieldWidth:145
            label replace_label "Replace:" pos:[15,90] width:45 height:15
            edittext replace_text "" pos:[60,88] fieldWidth:145
            checkbox match_case_checkbox "Match Case" pos:[15,110] width:80 height:15 checked:true
            checkbox first_only_checkbox "First Occurrence Only" pos:[15,130] width:125 height:15
            
        groupbox edit_group "3. Edit Name" pos:[5,155] width:215 height:80
            label prefix_label "Prefix:" pos:[15,175] width:30 height:15
            edittext prefix_text "" pos:[45,173] fieldWidth:160  
            label base_label "Base:" pos:[15,195] width:30 height:15
            edittext base_text "" pos:[45,193] fieldWidth:160
            label suffix_label "Suffix:" pos:[15,215] width:30 height:15
            edittext suffix_text "" pos:[45,213] fieldWidth:160
            
        groupbox increment_group "4. Increment" pos:[5,240] width:215 height:70
            radioButtons pad_mode_radio "" pos:[15,260] width:80 height:50 labels:#("Numbered", "Lettered", "None") default:3 columns:1
            spinner padding_spinner "Padding:" pos:[130,261] range:[1,20,2] fieldWidth:25 type:#integer scale:1 enabled:false
        
        groupbox misc_group "5. Misc" pos:[5,315] width:215 height:75
            label case_label "Casing:" pos:[15,335] width:35 height:15
            radiobuttons case_radio "" pos:[55,335] labels:#("No Change", "UPPERCASE", "lowercase") columns:2
            checkbox name_length_checkbox "Shorten Names More Than " pos:[15,370] width:150 height:15
            spinner name_length_spinner "" pos:[170,370] range:[10,999,60] fieldWidth:30 type:#integer scale:1
        
        groupbox option_group "Options" pos:[5,395] width:215 height:100
            checkbox include_sub_checkbox "Include Sub Materials" pos:[15,415] height:15
            checkbox unique_name_checkbox "Force Unique Names" pos:[15,435] height:15 checked:true
            checkbox reset_checkbox "Reset All Fields After Rename" pos:[15,455] height:15 checked:true
            checkbox preview_checkbox "Preview New Names (Print Only)" pos:[15,475] height:15
        
        MultiListBox material_list "" pos:[230,8] width:500 height:37
        label filter_label "Filter:" pos:[230,507] width:35 height:15
        edittext filter_text "" pos:[265,505] fieldWidth:160  
        button select_all_button "Select All" pos:[538,505] width:91 height:20
        button select_none_button "Select None" pos:[639,505] width:91 height:20
        
        button rename_button "Rename Selected Materials" pos:[5,505] width:215 height:20
        label status_label "INFO: Standing by..." pos:[5,530] width:730 height:15

        /**********************************************************************
        UI open event handler. Refreshes the material list
        **********************************************************************/
        on rename_material_rollout open do
            refresh_list()
        
        /**********************************************************************
        UI move event handler. Stores the curent position of the UI
        **********************************************************************/
        on rename_material_rollout moved pos do
            setIniSetting rename_material_ini "Dialog" "Position" (pos as string)
            
        /**********************************************************************
        Padding type changed event handler. Sets the numeric padding value enable state
        **********************************************************************/
        on pad_mode_radio changed state do
        (
            if state == 1 then
                padding_spinner.enabled = true
            else
                padding_spinner.enabled = false
        ) --end on
        
        /**********************************************************************
        Include sub materials changed event handler. Refreshes the material list
        to show/hide sub materials
        **********************************************************************/
        on include_sub_checkbox changed state do
        (
            refresh_list()
            
            if state then
                status_label.text = "INFO: Sub materials included"
            else
                status_label.text = "INFO: Sub materials hidden"
        ) --end on
        
        /**********************************************************************
        Preview rename changed event handler. Changes the text of the rename button
        **********************************************************************/
        on preview_checkbox changed state do
        (
            if state then
                rename_button.text = "Preview Selected Materials"
            else
                rename_button.text = "Rename Selected Materials"
        ) --end on
        
        /**********************************************************************
        Filter text changed event handler. Applies the new filter to the material list
        **********************************************************************/
        on filter_text changed new_filter do
        (
            renamer.filter_materials new_filter
            material_list.items = for current in renamer.materials_filtered collect current.name
            
            if new_filter != "" then
                status_label.text = "INFO: Materials filtered by \"" + new_filter + "\""
            else
                status_label.text = "INFO: Material filter removed"
        ) --end on
        
        /**********************************************************************
        Select all clicked event handler. Selects all materials
        **********************************************************************/
        on select_all_button pressed do
        (
            material_list.selection = #{1..material_list.items.count}
            status_label.text = "INFO: All materials selected"
        ) --end on
        
        /**********************************************************************
        Select none clicked event handler. Clears the material selection
        **********************************************************************/
        on select_none_button pressed do
        (
            material_list.selection = #{}
            status_label.text = "INFO: All materials deselected"
        ) --end on
        
        /**********************************************************************
        Rename button event handler. Renames the selected materials
        **********************************************************************/
        on rename_button pressed do
        (
            local material_selection = material_list.selection as array
            local new_name = ""
            local counter = 1
            local warning_counter = 0
            
            if material_selection.count == 0 do
            (
                status_label.text = "ERROR: No material(s) selected"
                return false
            ) --end if
            
            if find_text.text != "" do
            (
                if (findString replace_text.text find_text.text) != undefined and first_only_checkbox.checked == false do
                (
                    first_only_checkbox.checked = true
                    print "WARNING: When the \"Find\" text is in the \"Replace\" text only the first occurrence can be changed"
                    warning_counter += 1
                ) --end if
            ) --end if
            
            -- CRITICAL! You must set counter to 0 if alphabetic
            if pad_mode_radio.state == 2 do
                counter = 0
            
            -- CRITICAL! Reset preview names!
            if preview_checkbox.checked do
            (
                renamer.material_previews = for current in renamer.materials_filtered collect current.name
                print "***   BEGIN RENAME PREVIEW   ***"
            ) --end if
            
            for index in material_selection do
            (
                new_name = renamer.materials_filtered[index].name
                
                if remove_first_spinner.value > 0 or remove_last_spinner.value > 0 do
                    new_name = renamer.get_remove new_name remove_first_spinner.value remove_last_spinner.value
                    
                if find_text.text != "" do
                (
                    if match_case_checkbox.checked then
                        new_name = renamer.get_case_replace new_name find_text.text replace_text.text first_only_checkbox.checked
                    else
                        new_name = renamer.get_replace new_name find_text.text replace_text.text first_only_checkbox.checked
                ) --end if
                    
                if base_text.text != "" or prefix_text.text != "" or suffix_text.text != "" do
                    new_name = renamer.get_edit new_name prefix_text.text base_text.text suffix_text.text
                
                case pad_mode_radio.state of
                (
                    (1): new_name += (renamer.get_numeric_padding counter padding_spinner.value)
                    (2): new_name += (renamer.get_alpha_padding counter)
                ) --end if
                        
                counter += 1
                
                if case_radio.state > 1 do
                    new_name = renamer.get_case new_name case_radio.state
                    
                if name_length_checkbox.checked do
                    new_name = renamer.get_shorten new_name name_length_spinner.value
                        
                if new_name.count > 0 then
                (
                    if unique_name_checkbox.checked do
                        new_name = renamer.get_unique new_name index
                        
                    if preview_checkbox.checked then
                    (
                        print (renamer.materials_filtered[index].name + " -> " + new_name)
                        renamer.material_previews[index] = new_name
                    ) --end if
                    else
                        renamer.materials_filtered[index].name = new_name
                ) --end if
                else
                (
                    print ("ERROR: Unable to rename material " + renamer.materials_filtered[index].name + ". New name is \"\" (null)")
                    warning_counter += 1
                ) --end else
            ) --end for
            
            if preview_checkbox.checked then
            (
                print "***   END RENAME PREVIEW   ***"
                status_label.text = ("INFO: Rename preview printed to listener. Warnings: " + (warning_counter as string))
            ) --end if
            else
            (
                if reset_checkbox.checked do
                (
                    reset_remove_ui()
                    reset_find_ui()
                    reset_edit_ui()
                    reset_increment_ui()
                    reset_misc_ui()
                ) --end if
                
                refresh_list()
                material_list.selection = #{}
                
                if warning_counter == 0 then
                    status_label.text = "INFO: Materials renamed. Errors: 0"
                else
                    status_label.text = ("WARNING: Materials renamed. Errors: "  + (warning_counter as string) + ". Check listener for details")
            ) --end else
        ) --end on
    ) --end rollout
    
    rename_material_position = execute (getIniSetting rename_material_ini "Dialog" "Position")
    if rename_material_position == OK do rename_material_position = [100,100]
    if rename_material_position.x > sysinfo.desktopsize.x - 740 do
        rename_material_position.x = sysinfo.desktopsize.x - 745
    if rename_material_position.y > sysinfo.desktopsize.y - 605 do
        rename_material_position.y = sysinfo.desktopsize.y - 605
    if rename_material_position.x < 0 do rename_material_position.x = 0 
    if rename_material_position.y < 0 do rename_material_position.y = 0
    createDialog rename_material_rollout 740 545 rename_material_position.x rename_material_position.y
) --end local scope
